@@ -9,6 +9,7 @@
  * %End-Header%
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -92,6 +93,10 @@ void do_create_icount(int argc, char **argv)
 
 void do_free_icount(int argc, char **argv)
 {
+	if (argc != 1) {
+		printf("Usage: free_icount\n");
+		return;
+	}
 	if (check_icount(argv[0]))
 		return;
 
@@ -205,6 +210,10 @@ void do_dump(int argc, char **argv)
 	ext2_ino_t	i;
 	__u16		count;
 
+	if (argc != 1) {
+		printf("Usage: dump\n");
+		return;
+	}
 	if (check_icount(argv[0]))
 		return;
 	for (i=1; i <= test_fs->super->s_inodes_count; i++) {
@@ -223,6 +232,10 @@ void do_validate(int argc, char **argv)
 {
 	errcode_t	retval;
 
+	if (argc != 1) {
+		printf("Usage: validate\n");
+		return;
+	}
 	if (check_icount(argv[0]))
 		return;
 	retval = ext2fs_icount_validate(test_icount, stdout);
@@ -237,6 +250,10 @@ void do_get_size(int argc, char **argv)
 {
 	ext2_ino_t	size;
 
+	if (argc != 1) {
+		printf("Usage: get_size\n");
+		return;
+	}
 	if (check_icount(argv[0]))
 		return;
 	size = ext2fs_get_icount_size(test_icount);
@@ -289,6 +306,8 @@ static int source_file(const char *cmd_file, int sci_idx)
 			exit_status++;
 		}
 	}
+	if (f != stdin)
+		fclose(f);
 	return exit_status;
 }
 
@@ -296,7 +315,6 @@ int main(int argc, char **argv)
 {
 	int		retval;
 	int		sci_idx;
-	const char	*usage = "Usage: test_icount [-R request] [-f cmd_file]";
 	int		c;
 	char		*request = 0;
 	int		exit_status = 0;
@@ -309,7 +327,7 @@ int main(int argc, char **argv)
 	 * Create a sample filesystem structure
 	 */
 	memset(&param, 0, sizeof(struct ext2_super_block));
-	param.s_blocks_count = 80000;
+	ext2fs_blocks_count_set(&param, 80000);
 	param.s_inodes_count = 20000;
 	retval = ext2fs_initialize("/dev/null", 0, &param,
 				   unix_io_manager, &test_fs);
@@ -327,7 +345,8 @@ int main(int argc, char **argv)
 			cmd_file = optarg;
 			break;
 		default:
-			com_err(argv[0], 0, usage);
+			com_err(argv[0], 0, "Usage: test_icount "
+				"[-R request] [-f cmd_file]");
 			exit(1);
 		}
 	}
